==> mcloud-0.9.6.tar.gz: mcloud-0.9.6/mcloud/plugins/haproxy.py <==
import logging
import inject
from mcloud.application import ApplicationController
from mcloud.events import EventBus
from mcloud.plugins import Plugin
import os
from twisted.internet.defer import inlineCallbacks
from twisted.python import log

HAPROXY_TPL = """
global
        log /dev/log    local0
        log /dev/log    local1 notice
        chroot /var/lib/haproxy
        user haproxy
        group haproxy
        daemon

defaults
        log     global
        option  dontlognull
        timeout connect 5000
        timeout client 50000
        timeout server 50000

        errorfile 400 /etc/haproxy/errors/400.http
        errorfile 403 /etc/haproxy/errors/403.http
        errorfile 408 /etc/haproxy/errors/408.http
        errorfile 500 /etc/haproxy/errors/500.http
        errorfile 502 /etc/haproxy/errors/502.http
        errorfile 503 /etc/haproxy/errors/503.http
        errorfile 504 /etc/haproxy/errors/504.http


{% if ssl_apps %}
frontend http_ssl_proxy
  mode tcp
  bind 0.0.0.0:443

  tcp-request inspect-delay 5s
  tcp-request content accept if { req_ssl_hello_type 1 }

  {% for app in ssl_apps %}
  {% for domain in app.domains %}
  acl is_ssl_{{ app.name }} req_ssl_sni -i {{ domain }}
  {% endfor %}
  use_backend backend_ssl_{{ app.name }}_cluster if is_ssl_{{ app.name }}
  {% endfor %}

  {% for app in ssl_apps %}
  {% for backend in app.backends %}
  backend {{ backend.name }}_cluster
      mode tcp

      # maximum SSL session ID length is 32 bytes.
      stick-table type binary len 32 size 30k expire 30m

      acl clienthello req_ssl_hello_type 1
      acl serverhello rep_ssl_hello_type 2

      # use tcp content accepts to detects ssl client and server hello.
      tcp-request inspect-delay 5s
      tcp-request content accept if clienthello

      # no timeout on response inspect delay by default.
      tcp-response content accept if serverhello

      stick on payload_lv(43,1) if clienthello

      # Learn on response if server hello.
      stick store-response payload_lv(43,1) if serverhello

      option ssl-hello-chk

      server {{ backend.name }} {{ backend.ip }}:{{ backend.port }} check

  {% endfor %}
  {% endfor %}
{% endif %}

frontend http_proxy
  bind 0.0.0.0:80

  mode    http
  option  httplog
  option  httpclose
  option  forwardfor

  {% for app in apps %}
  {% for domain in app.domains %}
  acl is_{{ app.name }} hdr(host) -i {{ domain }}
  {% endfor %}
  use_backend backend_{{ app.name }}_cluster if is_{{ app.name }}
  {% endfor %}

  {% for app in apps %}
  {% for backend in app.backends %}
  backend {{ backend.name }}_cluster
      mode    http
      server {{ backend.name }} {{ backend.ip }}:{{ backend.port }}
  {% endfor %}
  {% endfor %}
"""

from jinja2 import Template

logger = logging.getLogger('mcloud.plugin.haproxy')

class HaproxyConfig(object):

    def __init__(self, path, template=None):
        self.template = template
        self.path = path

    def dump(self, apps_list):

        template = self.template

        if not template:
            template = Template(HAPROXY_TPL)

        proxy_apps = []
        proxy_ssl_apps = []

        for app in apps_list:
            # if not 'web_target' in app or not app['web_target']:
            #     continue

            plain_domains = {app['web_target']: [app['fullname']]}
            ssl_domains = {}

            print app['fullname']

            if app['public_urls']:
                for target in app['public_urls']:

                    if not target['service']:
                        if 'ssl_target' in app and app['ssl_target'] and target['url'].startswith('https://'):
                            if not app['ssl_target'] in ssl_domains:
                                ssl_domains[app['ssl_target']] = []
                            ssl_domains[app['ssl_target']].append(target['url'][8:])

                        if 'web_target' in app and app['web_target'] and not target['url'].startswith('https://'):
                            if not app['web_target'] in plain_domains:
                                plain_domains[app['web_target']] = []
                            plain_domains[app['web_target']].append(target['url'])

                    else:
                        for service in app['services']:
                            if service['shortname'] == target['service']:

                                print target

                                if 'port' in target and target['port']:
                                    service['ip'] = service['ip'] + ':' + target['port']


                                if target['url'].startswith('https://'):
                                    if not service['ip'] in ssl_domains:
                                        ssl_domains[service['ip']] = []
                                    ssl_domains[service['ip']].append(target['url'][8:])
                                else:
                                    if not service['ip'] in plain_domains:
                                        plain_domains[service['ip']] = []
                                    plain_domains[service['ip']].append(target['url'])

            if ssl_domains:
                for ip, domains in ssl_domains.items():
                    port = 443
                    if ':' in ip:
                        ip, port = ip.split(':')

                    proxy_ssl_apps.append({
                        'name': '%s_%s_%s' % (app['fullname'], ip.replace('.', '_'), port),
                        'domains': domains,
                        'backends': [{'name': 'backend_ssl_%s_%s_%s' % (app['fullname'], ip.replace('.', '_'), port), 'ip': ip, 'port': port}]
                    })

            for ip, domains in plain_domains.items():
                if ip is None:
                    continue

                port = 80
                if ':' in ip:
                    ip, port = ip.split(':')

                proxy_apps.append({
                    'name': '%s_%s_%s' % (app['fullname'], ip.replace('.', '_'), port),
                    'domains': domains,
                    'backends': [{'name': 'backend_%s_%s_%s' % (app['fullname'], ip.replace('.', '_'), port), 'ip': ip, 'port': port}]
                })

        log.msg('Writing haproxy config')

        with open('/etc/haproxy/haproxy.cfg', 'w') as f:
            f.write(template.render({'apps': proxy_apps, 'ssl_apps': proxy_ssl_apps}))

        os.system('service haproxy reload')


class HaproxyPlugin(Plugin):
    eb = inject.attr(EventBus)
    app_controller = inject.attr(ApplicationController)

    def __init__(self):
        super(HaproxyPlugin, self).__init__()

        self.eb.on('containers.updated', self.containers_updated)
        self.proxy_config = HaproxyConfig(path='/etc/haproxy/haproxy.cfg')

        logger.info('Haproxy plugin started')

        self.containers_updated()

    @inlineCallbacks
    def containers_updated(self, *args, **kwargs):
        logger.info('Containers updated: dumping haproxy config.')
        data = yield self.app_controller.list()
        self.proxy_config.dump(data)

==> mcloud-0.9.11.tar.gz: mcloud-0.9.11/mcloud/plugins/haproxy.py <==
import logging
import datetime
from autobahn.twisted.util import sleep
import inject
from mcloud.application import ApplicationController
from mcloud.container import PrebuiltImageBuilder, InlineDockerfileImageBuilder
from mcloud.events import EventBus
from mcloud.plugins import Plugin, PluginInitError
from mcloud.service import Service
import os
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.python import log

HAPROXY_TPL = """
global
        log 127.0.0.1 local0
        log 127.0.0.1 local1 notice
        chroot /var/lib/haproxy
        user haproxy
        group haproxy

defaults
        log     global
        option  dontlognull
        timeout connect 5000
        timeout client 50000
        timeout server 50000


{% if ssl_apps %}
frontend http_ssl_proxy
  mode tcp
  bind 0.0.0.0:443

  tcp-request inspect-delay 5s
  tcp-request content accept if { req_ssl_hello_type 1 }

  {% for app in ssl_apps %}
  {% for domain in app.domains %}
  acl is_ssl_{{ app.name }} req_ssl_sni -i {{ domain }}
  {% endfor %}
  use_backend backend_ssl_{{ app.name }}_cluster if is_ssl_{{ app.name }}
  {% endfor %}

  {% for app in ssl_apps %}
  {% for backend in app.backends %}
  backend {{ backend.name }}_cluster
      mode tcp

      # maximum SSL session ID length is 32 bytes.
      stick-table type binary len 32 size 30k expire 30m

      acl clienthello req_ssl_hello_type 1
      acl serverhello rep_ssl_hello_type 2

      # use tcp content accepts to detects ssl client and server hello.
      tcp-request inspect-delay 5s
      tcp-request content accept if clienthello

      # no timeout on response inspect delay by default.
      tcp-response content accept if serverhello

      stick on payload_lv(43,1) if clienthello

      # Learn on response if server hello.
      stick store-response payload_lv(43,1) if serverhello

      option ssl-hello-chk

      server {{ backend.name }} {{ backend.ip }}:{{ backend.port }} check

  {% endfor %}
  {% endfor %}
{% endif %}

frontend http_proxy
  bind 0.0.0.0:80

  mode    http
  option  httplog
  option  httpclose
  option  forwardfor

  {% for app in apps %}
  {% for domain in app.domains %}
  acl is_{{ app.name }} hdr(host) -i {{ domain }}
  {% endfor %}
  use_backend backend_{{ app.name }}_cluster if is_{{ app.name }}
  {% endfor %}

  {% for app in apps %}
  {% for backend in app.backends %}
  backend {{ backend.name }}_cluster
      mode    http
      server {{ backend.name }} {{ backend.ip }}:{{ backend.port }}
  {% endfor %}
  {% endfor %}
"""

from jinja2 import Template

logger = logging.getLogger('mcloud.plugin.haproxy')

class HaproxyConfig(object):

    def __init__(self, path, template=None):
        self.template = template
        self.path = path

    def dump(self, apps_list):

        template = self.template

        if not template:
            template = Template(HAPROXY_TPL)

        proxy_apps = []
        proxy_ssl_apps = []

        for app in apps_list:
            # if not 'web_target' in app or not app['web_target']:
            #     continue

            plain_domains = {app['web_target']: [app['fullname']]}
            ssl_domains = {}

            print app['fullname']

            if app['public_urls']:
                for target in app['public_urls']:

                    if not target['service']:
                        if 'ssl_target' in app and app['ssl_target'] and target['url'].startswith('https://'):
                            if not app['ssl_target'] in ssl_domains:
                                ssl_domains[app['ssl_target']] = []
                            ssl_domains[app['ssl_target']].append(target['url'][8:])

                        if 'web_target' in app and app['web_target'] and not target['url'].startswith('https://'):
                            if not app['web_target'] in plain_domains:
                                plain_domains[app['web_target']] = []
                            plain_domains[app['web_target']].append(target['url'])

                    else:
                        for service in app['services']:
                            if service['shortname'] == target['service']:

                                print target

                                if 'port' in target and target['port']:
                                    service['ip'] = service['ip'] + ':' + target['port']


                                if target['url'].startswith('https://'):
                                    if not service['ip'] in ssl_domains:
                                        ssl_domains[service['ip']] = []
                                    ssl_domains[service['ip']].append(target['url'][8:])
                                else:
                                    if not service['ip'] in plain_domains:
                                        plain_domains[service['ip']] = []
                                    plain_domains[service['ip']].append(target['url'])

            if ssl_domains:
                for ip, domains in ssl_domains.items():
                    port = 443
                    if ':' in ip:
                        ip, port = ip.split(':')

                    proxy_ssl_apps.append({
                        'name': '%s_%s_%s' % (app['fullname'], ip.replace('.', '_'), port),
                        'domains': domains,
                        'backends': [{'name': 'backend_ssl_%s_%s_%s' % (app['fullname'], ip.replace('.', '_'), port), 'ip': ip, 'port': port}]
                    })

            for ip, domains in plain_domains.items():
                if ip is None:
                    continue

                port = 80
                if ':' in ip:
                    ip, port = ip.split(':')

                proxy_apps.append({
                    'name': '%s_%s_%s' % (app['fullname'], ip.replace('.', '_'), port),
                    'domains': domains,
                    'backends': [{'name': 'backend_%s_%s_%s' % (app['fullname'], ip.replace('.', '_'), port), 'ip': ip, 'port': port}]
                })

        log.msg('Writing haproxy config')

        with open(self.path, 'w') as f:
            f.write(template.render({'apps': proxy_apps, 'ssl_apps': proxy_ssl_apps}))


class HaproxyPlugin(Plugin):
    eb = inject.attr(EventBus)
    settings = inject.attr('settings')
    app_controller = inject.attr(ApplicationController)

    last_reboot = None

    @inlineCallbacks
    def setup(self):

        haproxy_path = os.path.expanduser('%s/haproxy' % self.settings.home_dir)
        if not os.path.exists(haproxy_path):
            os.makedirs(haproxy_path)

        template_path = os.path.join(haproxy_path, 'haproxy.tpl')
        haproxy_config_path = os.path.join(haproxy_path, 'haproxy.cfg')

        if not os.path.exists(template_path):
            with open(template_path, 'w+') as f:
                f.write(HAPROXY_TPL)

        with open(template_path) as f:
            self.proxy_config = HaproxyConfig(path=haproxy_config_path, template=Template(f.read()))

        self.haproxy = Service()
        self.haproxy.name = 'mcloud_haproxy'
        self.haproxy.image_builder = InlineDockerfileImageBuilder(source="""

        FROM ubuntu

        RUN apt-key adv --keyserver keyserver.ubuntu.com --recv-keys 1C61B9CD
        RUN echo "deb http://ppa.launchpad.net/vbernat/haproxy-1.5/ubuntu trusty main" >> /etc/apt/sources.list.d/haproxy.list

        RUN apt-get update && apt-get install -y haproxy
        RUN sed -i 's/^ENABLED=.*/ENABLED=1/' /etc/default/haproxy && \
          rm -rf /var/lib/apt/lists/*

        WORKDIR /etc/haproxy

        # Define default command.
        CMD ["haproxy", "-f", "/etc/haproxy/haproxy.cfg"]

        # Expose ports.
        EXPOSE 80
        EXPOSE 443

        """)
        self.haproxy.ports = ['80/tcp:80', '443/tcp:443']
        self.haproxy.volumes = [{
            'local': haproxy_path,
            'remote': '/etc/haproxy'
        }]

        yield self.containers_updated()

        # subscribe to container events
        self.eb.on('containers.updated', self.containers_updated)

        logger.info('Haproxy plugin started')

    def can_reboot(self):
        if not self.last_reboot:
            self.last_reboot = datetime.datetime.now()
            return True

        now = datetime.datetime.now()
        if (now - self.last_reboot).seconds < 5:
            return False
        self.last_reboot = datetime.datetime.now()
        return True

    @inlineCallbacks
    def containers_updated(self, *args, **kwargs):
        if not self.can_reboot():
            return

        logger.info('Containers updated: dumping haproxy config.')
        data = yield self.app_controller.list()
        self.proxy_config.dump(data)

        yield self.haproxy.restart()
